@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
 # Copyright 2010 Fabian Moser
-# Copyright 2011 Jaap Karssenberg
+# Copyright 2011-2014 Jaap Karssenberg
 
 
 import gobject
@@ -10,18 +10,201 @@ import pango
 
 import logging
 
-from zim.plugins import PluginClass
+from zim.plugins import PluginClass, extends, WindowExtension
 from zim.gui.pageindex import PageTreeStore, PageTreeIter, PageTreeView, \
-	NAME_COL, PATH_COL, EMPTY_COL, STYLE_COL, FGCOLOR_COL, WEIGHT_COL, N_CHILD_COL
+	NAME_COL, PATH_COL, EMPTY_COL, STYLE_COL, FGCOLOR_COL, WEIGHT_COL, N_CHILD_COL, TIP_COL
 from zim.notebook import Path
 from zim.index import IndexPath, IndexTag
-from zim.gui.widgets import LEFT_PANE, populate_popup_add_separator
+from zim.gui.widgets import LEFT_PANE, PANE_POSITIONS, populate_popup_add_separator, ScrolledWindow, encode_markup_text
 from zim.gui.clipboard import pack_urilist, INTERNAL_PAGELIST_TARGET_NAME
+from zim.signals import ConnectorMixin
 
 
 logger = logging.getLogger('zim.plugins.tags')
 
 
+
+class TagsPlugin(PluginClass):
+
+	plugin_info = {
+		'name': _('Tags'), # T: plugin name
+		'description': _('''\
+This plugin provides a page index filtered by means of selecting tags in a cloud.
+'''), # T: plugin description
+		'author': 'Fabian Moser',
+		'help': 'Plugins:Tags',
+	}
+
+	plugin_preferences = (
+		# key, type, label, default
+		('pane', 'choice', _('Position in the window'), LEFT_PANE, PANE_POSITIONS),
+			# T: option for plugin preferences
+	)
+
+
+@extends('MainWindow')
+class MainWindowExtension(WindowExtension):
+
+	def __init__(self, plugin, window):
+		WindowExtension.__init__(self, plugin, window)
+
+		self.widget = TagsPluginWidget(self.window.ui.notebook.index, self.uistate, self.window.ui) # XXX
+
+		self.on_preferences_changed(plugin.preferences)
+		self.connectto(plugin.preferences, 'changed', self.on_preferences_changed)
+
+		self.uistate.setdefault('vpane_pos', 150)
+		self.widget.set_position(self.uistate['vpane_pos'])
+		def update_uistate(*a):
+			self.uistate['vpane_pos'] = self.widget.get_position()
+		self.widget.connect('notify::position', update_uistate)
+
+	def on_preferences_changed(self, preferences):
+		if self.widget is None:
+			return
+
+		try:
+			self.window.remove(self.widget)
+		except ValueError:
+			pass
+		self.window.add_tab(_('Tags'), self.widget, preferences['pane'])
+		self.widget.show_all()
+
+	def teardown(self):
+		self.window.remove(self.widget)
+		self.widget.disconnect_all()
+		self.widget = None
+
+
+class TagsPluginWidget(ConnectorMixin, gtk.VPaned):
+	'''Widget combining a tag cloud and a tag based page treeview'''
+
+	def __init__(self, index, uistate, ui): # XXX
+		gtk.VPaned.__init__(self)
+		self.index = index
+		self.uistate = uistate
+
+		self.uistate.setdefault('treeview', 'tagged', set(['tagged', 'tags']))
+		self.uistate.setdefault('tagcloud_sorting', 'score', set(['alpha', 'score']))
+		self.uistate.setdefault('show_full_page_name', True)
+
+		self.tagcloud = TagCloudWidget(self.index, sorting=self.uistate['tagcloud_sorting'])
+		self.pack1(ScrolledWindow(self.tagcloud), shrink=False)
+
+		self.treeview = TagsPageTreeView(ui) # XXX
+		self._treeview_mode = (None, None)
+		self.pack2(ScrolledWindow(self.treeview), shrink=False)
+
+		self.treeview.connect('populate-popup', self.on_populate_popup)
+		self.tagcloud.connect('selection-changed', self.on_cloud_selection_changed)
+		self.tagcloud.connect('sorting-changed', self.on_cloud_sortin_changed)
+
+		self.connectto_all(ui, ( # XXX
+			'open-page',
+			('start-index-update', lambda o: self.disconnect_model()),
+			('end-index-update', lambda o: self.reload_model()),
+		))
+
+		self.reload_model()
+
+	def on_open_page(self, ui, page, path):
+		expand = True
+		treepath = self.treeview.set_current_page(path, vivificate=True)
+		expand = ui.notebook.namespace_properties[path.name].get('auto_expand_in_index', True)
+		if treepath and expand:
+			# change selection only if necessary
+			selected_path = self.treeview.get_selected_path()
+			if path == selected_path:
+				logger.debug('Already selected: "%s"', path)
+			else:
+				self.treeview.select_treepath(treepath)
+
+	def toggle_treeview(self):
+		'''Toggle the treeview type in the widget'''
+		if self.uistate['treeview'] == 'tagged':
+			self.uistate['treeview'] = 'tags'
+		else:
+			self.uistate['treeview'] = 'tagged'
+
+		self.reload_model()
+
+	def toggle_show_full_page_name(self):
+		self.uistate['show_full_page_name'] = not self.uistate['show_full_page_name']
+		self.reload_model()
+
+	def on_populate_popup(self, treeview, menu):
+		# If not a not a page (but e.g. a tag), remove page action
+		if not treeview.get_selected_path():
+			for item in menu.get_children():
+				menu.remove(item)
+
+		# Add a popup menu item to switch the treeview mode
+		populate_popup_add_separator(menu, prepend=True)
+
+		item = gtk.CheckMenuItem(_('Show full page name')) # T: menu option
+		item.set_active(self.uistate['show_full_page_name'])
+		item.connect_object('toggled', self.__class__.toggle_show_full_page_name, self)
+		menu.prepend(item)
+
+		item = gtk.CheckMenuItem(_('Sort pages by tags')) # T: menu option
+		item.set_active(self.uistate['treeview'] == 'tags')
+		item.connect_object('toggled', self.__class__.toggle_treeview, self)
+		menu.prepend(item)
+
+		menu.show_all()
+
+	def on_cloud_selection_changed(self, cloud):
+		filter = cloud.get_tag_filter()
+		type, was_filtered = self._treeview_mode
+		is_filtered = (filter is not None)
+		if type == 'tagged' and was_filtered != is_filtered:
+			# Switch between tag view and normal index or vice versa
+			self._reload_model(type, filter)
+		else:
+			self.treeview.set_tag_filter(filter)
+
+	def on_cloud_sortin_changed(self, cloud, sorting):
+		self.uistate['tagcloud_sorting'] = sorting
+
+	def disconnect_model(self):
+		'''Stop the model from listening to the index. Used to
+		unhook the model before reloading the index. Typically
+		should be followed by reload_model().
+		'''
+		self.treeview.disconnect_index()
+		self.tagcloud.disconnect_index()
+
+	def reload_model(self):
+		'''Re-initialize the treeview model. This is called when
+		reloading the index to get rid of out-of-sync model errors
+		without need to close the app first.
+		'''
+		assert self.uistate['treeview'] in ('tagged', 'tags')
+
+		if self.tagcloud.index is None:
+			self.tagcloud.set_index(self.index)
+
+		type = self.uistate['treeview']
+		filter = self.tagcloud.get_tag_filter()
+		self._reload_model(type, filter)
+
+	def _reload_model(self, type, filter):
+		if type == 'tagged':
+			if filter is None:
+				model = DuplicatePageTreeStore(self.index)
+					# show the normal index in this case
+			else:
+				model = TaggedPageTreeStore(self.index, self.uistate['show_full_page_name'])
+		elif type == 'tags':
+			model = TagsPageTreeStore(self.index, self.uistate['show_full_page_name'])
+		else:
+			assert False
+
+		is_filtered = (filter is not None)
+		self._treeview_mode = (type, is_filtered)
+		self.treeview.set_model(model, filter)
+
+
 class PageTreeTagIter(object):
 	'''Simple wrapper for IndexTag objects used as tree iters
 
@@ -46,15 +229,17 @@ class DuplicatePageTreeStore(PageTreeStore):
 	multiple times in the tree.
 	'''
 
-	def select_page(self, path):
-		oldpath = self.selected_page
-		self.selected_page = path
+	def set_current_page(self, path):
+		'''Since there may be duplicates of each page, highlight all of them'''
+		oldpath = self.current_page
+		self.current_page = path
 
 		for mypath in (oldpath, path):
 			if mypath:
 				for treepath in self.get_treepaths(mypath):
-					treeiter = self.get_iter(treepath)
-					self.emit('row-changed', treepath, treeiter)
+					if treepath:
+						treeiter = self.get_iter(treepath)
+						self.emit('row-changed', treepath, treeiter)
 
 	def get_treepath(self, path):
 		# Just returns the first treepath matching notebook path
@@ -65,8 +250,13 @@ class DuplicatePageTreeStore(PageTreeStore):
 			return None
 
 	def get_treepaths(self, path):
-		'''Return all treepaths matching notebook path 'path' '''
-		raise NotImplementedError
+		'''Return all treepaths matching notebook path 'path'
+		Default implementation assumes we are a non-duplicate treeview
+		after all and uses L{PageTreeStore.get_treepath()}.
+		@implementation: must be overloaded by subclasses that are real
+		duplicate stores
+		'''
+		return [PageTreeStore.get_treepath(self, path)]
 
 
 class TagsPageTreeStore(DuplicatePageTreeStore):
@@ -79,8 +269,9 @@ class TagsPageTreeStore(DuplicatePageTreeStore):
 
 	filter_depth = 2 # tag filter applies to top two levels
 
-	def __init__(self, index):
+	def __init__(self, index, show_full_page_name=True):
 		self._reverse_cache = {}
+		self.show_full_page_name = show_full_page_name
 		self.untagged = IndexTag(_('untagged'), -1)
 			# T: label for untagged pages in side pane
 		PageTreeStore.__init__(self, index)
@@ -92,8 +283,12 @@ class TagsPageTreeStore(DuplicatePageTreeStore):
 			treepaths = self.get_treepaths(path)
 			for treepath in treepaths:
 				#~ print '!!', signal, path, treepath
-				treeiter = self.get_iter(treepath)
-				self.emit(signal, treepath, treeiter)
+				try:
+					treeiter = self.get_iter(treepath)
+				except:
+					logger.exception('BUG: Invalid treepath: %s %s %s', signal, path, treepath)
+				else:
+					self.emit(signal, treepath, treeiter)
 
 		def on_page_deleted(o, path):
 			#~ print '!! page delete', path
@@ -164,19 +359,19 @@ class TagsPageTreeStore(DuplicatePageTreeStore):
 			self.row_deleted(treepath)
 			self._flush()
 
-		self._signals = (
-			self.index.connect('page-inserted', on_page_changed, 'row-inserted'),
-			self.index.connect('page-updated', on_page_changed, 'row-changed'),
-			self.index.connect('page-haschildren-toggled', on_page_changed, 'row-has-child-toggled'),
-			self.index.connect('page-to-be-deleted', on_page_deleted),
+		self.connectto_all(self.index, (
+			('page-inserted', on_page_changed, 'row-inserted'),
+			('page-updated', on_page_changed, 'row-changed'),
+			('page-haschildren-toggled', on_page_changed, 'row-has-child-toggled'),
+			('page-to-be-deleted', on_page_deleted),
 			# TODO: Treat tag-inserted and new tag differently
-			self.index.connect('tag-created', on_tag_created),
-			self.index.connect('tag-to-be-inserted', on_tag_to_be_inserted),
-			self.index.connect('tag-inserted', on_tag_inserted),
-			self.index.connect('tag-to-be-removed', on_tag_to_be_removed),
-			self.index.connect('tag-removed', on_tag_removed),
-			self.index.connect('tag-to-be-deleted', on_tag_to_be_deleted),
-		)
+			('tag-created', on_tag_created),
+			('tag-to-be-inserted', on_tag_to_be_inserted),
+			('tag-inserted', on_tag_inserted),
+			('tag-to-be-removed', on_tag_to_be_removed),
+			('tag-removed', on_tag_removed),
+			('tag-to-be-deleted', on_tag_to_be_deleted),
+		))
 		# The page-to-be-deleted signal is a hack so we have time to ensure we know the
 		# treepath of this indexpath - once we get page-deleted it is to late to get this
 
@@ -364,6 +559,8 @@ class TagsPageTreeStore(DuplicatePageTreeStore):
 			tag = iter.indextag
 			if column == NAME_COL:
 				return tag.name
+			elif column == TIP_COL:
+				return encode_markup_text(tag.name)
 			elif column == PATH_COL:
 				return tag
 			elif column == EMPTY_COL:
@@ -389,7 +586,15 @@ class TagsPageTreeStore(DuplicatePageTreeStore):
 				#~ else:
 					#~ return str(self.index.n_list_tagged_pages(tag))
 		else:
-			return PageTreeStore.on_get_value(self, iter, column)
+			if column == NAME_COL and self.show_full_page_name:
+				# Show top level pages with full contex
+				# top level tree is tags, so top level pages len(path) is 2
+				if len(iter.treepath) <= 2:
+					return iter.indexpath.name
+				else:
+					return iter.indexpath.basename
+			else:
+				return PageTreeStore.on_get_value(self, iter, column)
 
 
 class TaggedPageTreeStore(DuplicatePageTreeStore):
@@ -400,9 +605,10 @@ class TaggedPageTreeStore(DuplicatePageTreeStore):
 
 	filter_depth = 1 # tag filter only applies to top level
 
-	def __init__(self, index):
+	def __init__(self, index, show_full_page_name=True):
 		PageTreeStore.__init__(self, index)
 		self._reverse_cache = {}
+		self.show_full_page_name = show_full_page_name
 
 	def _connect(self):
 		def on_page_changed(o, path, signal):
@@ -418,19 +624,19 @@ class TaggedPageTreeStore(DuplicatePageTreeStore):
 				self.emit('row-deleted', treepath)
 			self._flush()
 
-		self._signals = (
-			self.index.connect('page-inserted', on_page_changed, 'row-inserted'),
-			self.index.connect('page-updated', on_page_changed, 'row-changed'),
-			self.index.connect('page-haschildren-toggled', on_page_changed, 'row-has-child-toggled'),
-			self.index.connect('page-to-be-deleted', on_page_deleted),
-		)
+		self.connectto_all(self.index, (
+			('page-inserted', on_page_changed, 'row-inserted'),
+			('page-updated', on_page_changed, 'row-changed'),
+			('page-haschildren-toggled', on_page_changed, 'row-has-child-toggled'),
+			('page-to-be-deleted', on_page_deleted),
+		))
 
 	def _get_iter(self, treepath):
 		'''
 		Cached conversion of the tree hierarchy to a PageTreeIter.
 
 		@param treepath: A tuple of int e.g. (0,) is the first item in the root namespace.
-		@return: A PageTreeIter instance corresponding to the given path
+		@returns: A PageTreeIter instance corresponding to the given path
 		'''
 		if not treepath in self._cache:
 			parent = None
@@ -470,7 +676,7 @@ class TaggedPageTreeStore(DuplicatePageTreeStore):
 		the inverse operation of _get_iter.
 
 		@param path: Usually an IndexPath instance
-		@return: A list of tuples of ints (one page can be represented many times)
+		@returns: A list of tuples of ints (one page can be represented many times)
 		'''
 		assert isinstance(path, Path)
 		if path.isroot:
@@ -511,6 +717,17 @@ class TaggedPageTreeStore(DuplicatePageTreeStore):
 		else:
 			return PageTreeStore.on_iter_n_children(self, iter)
 
+	def on_get_value(self, iter, column):
+		'''Returns the data for a specific column'''
+		if column == NAME_COL and self.show_full_page_name:
+			# Show top level pages with full contex
+			if len(iter.treepath) == 1:
+				return iter.indexpath.name
+			else:
+				return iter.indexpath.basename
+		else:
+			return PageTreeStore.on_get_value(self, iter, column)
+
 
 class TagsPageTreeView(PageTreeView):
 
@@ -522,7 +739,7 @@ class TagsPageTreeView(PageTreeView):
 		if model:
 			self.set_model(model)
 
-	def set_model(self, model):
+	def set_model(self, model, filter=None):
 		'''Set the model to be used'''
 		# disconnect previous model
 		oldmodel = self.get_model()
@@ -530,6 +747,10 @@ class TagsPageTreeView(PageTreeView):
 			childmodel = oldmodel.get_model()
 			childmodel.disconnect_index()
 
+		# Filter is also provided here, just to make it more efficient to
+		# set model and filter in one go without need for refilter
+		self._set_tag_filter(filter)
+
 		# set new model
 		def func(model, iter):
 			index = self.ui.notebook.index
@@ -552,8 +773,11 @@ class TagsPageTreeView(PageTreeView):
 		# HACK add some methods and attributes
 		# (can not subclass gtk.TreeModelFilter because it lacks a constructor)
 		def get_indexpath(treeiter):
-			return model.get_indexpath(
-				filtermodel.convert_iter_to_child_iter(treeiter) )
+			childiter = filtermodel.convert_iter_to_child_iter(treeiter)
+			if childiter:
+				return model.get_indexpath(childiter)
+			else:
+				return None
 
 		def get_treepath(path):
 			for treepath in model.get_treepaths(path):
@@ -575,23 +799,31 @@ class TagsPageTreeView(PageTreeView):
 		filtermodel.get_treepath = get_treepath
 		filtermodel.get_treepaths = get_treepaths
 		filtermodel.index = model.index
-		filtermodel.select_page = model.select_page
+		filtermodel.set_current_page = model.set_current_page
 
 		PageTreeView.set_model(self, filtermodel)
 
 	def set_tag_filter(self, filter):
 		'''Sets the tags to filter on. The filter should be a tuple of
 		two lists of tags, or None to not do any filtering.
+		First list of tags are the tags that we filter on, so only pages
+		matching all these tags should be selected.
+		Second set is a superset of the first set and includes all tags
+		that appear in one of the selected pages. So selecting one of these
+		tags on top of the current selection should result in a subset
+		of the current page selection.
 		'''
+		self._set_tag_filter(filter)
+		model = self.get_model()
+		if model:
+			model.refilter()
+
+	def _set_tag_filter(self, filter):
 		if not filter:
 			self._tag_filter = None
 		else:
 			self._tag_filter = (frozenset(filter[0]), frozenset(filter[1]))
 
-		model = self.get_model()
-		if model:
-			model.refilter()
-
 	def do_drag_data_get(self, dragcontext, selectiondata, info, time):
 		assert selectiondata.target == INTERNAL_PAGELIST_TARGET_NAME
 		model, iter = self.get_selection().get_selected()
@@ -604,6 +836,50 @@ class TagsPageTreeView(PageTreeView):
 		data = pack_urilist((link,))
 		selectiondata.set(INTERNAL_PAGELIST_TARGET_NAME, 8, data)
 
+	def set_current_page(self, path, vivificate=False):
+		'''Set the current page in the treeview
+
+		@param path: a notebook L{Path} object for the page
+		@keyword vivificate: when C{True} the path is created
+		temporarily when it did not yet exist
+
+		@returns: a gtk TreePath (tuple of intergers) or C{None}
+		'''
+		#~ print '!! SELECT', path
+		model = self.get_model()
+		if model is None:
+			return None # index not yet initialized ...
+
+		treepath = model.get_treepath(path)
+		if not treepath:
+			if vivificate:
+				# path does not exist, but we can create it
+				path = model.index.touch(path)
+				treepath = model.get_treepath(path)
+				if not treepath \
+				and isinstance(model, gtk.TreeModelFilter):
+					return None # path is filtered
+				assert treepath, 'BUG: failed to touch placeholder'
+			else:
+				# path does not exist and we are not going to create it
+				return None
+
+		rowreference = gtk.TreeRowReference(model, treepath)
+			# make reference before cleanup - path may change
+
+		if self._cleanup and self._cleanup.valid():
+			mytreepath = self._cleanup.get_path()
+			if mytreepath != treepath:
+				indexpath = model.get_indexpath( model.get_iter(mytreepath) )
+				#~ print '!! CLEANUP', indexpath
+				model.index.cleanup(indexpath)
+
+		self._cleanup = rowreference
+
+		model.set_current_page(path) # highlight in model
+
+		return treepath
+
 # Need to register classes defining gobject signals
 gobject.type_register(TagsPageTreeView)
 
@@ -612,7 +888,7 @@ class TagCloudItem(gtk.ToggleButton):
 	'''Button item used on the tag cloud widget'''
 
 	def __init__(self, indextag):
-		gtk.ToggleButton.__init__(self, indextag.name)
+		gtk.ToggleButton.__init__(self, indextag.name, use_underline=False)
 		self.set_relief(gtk.RELIEF_NONE)
 		self.indextag = indextag
 
@@ -628,7 +904,7 @@ class TagCloudItem(gtk.ToggleButton):
 		self.connect_after('toggled', update_label)
 
 
-class TagCloudWidget(gtk.TextView):
+class TagCloudWidget(ConnectorMixin, gtk.TextView):
 	'''Text-view based list of tags, where each tag is represented by a
 	button inserted as a child in the textview.
 
@@ -642,29 +918,26 @@ class TagCloudWidget(gtk.TextView):
 		'sorting-changed': (gobject.SIGNAL_RUN_LAST, None, (object,)),
 	}
 
-	def __init__(self, index=None, sorting='score'):
+	def __init__(self, index, sorting='score'):
 		gtk.TextView.__init__(self, None) # Create TextBuffer implicitly
 		self.set_name('zim-tags-tagcloud')
 		self.index = None
-		self._signals = ()
 
 		self.set_editable(False)
 		self.set_cursor_visible(False)
 		self.set_wrap_mode(gtk.WRAP_CHAR)
 
 		self.set_sorting(sorting)
-
-		if index:
-			self.set_index(index)
+		self.set_index(index)
 
 	def set_index(self, index):
 		'''Connect to an Index object'''
 		self.disconnect_index() # just to be sure
 		self.index = index
-		self._signals = (
-			self.index.connect('tag-created', self._update),
-			self.index.connect('tag-deleted', self._update),
-		)
+		self.connectto_all(self.index, (
+			('tag-created', self._update),
+			('tag-deleted', self._update),
+		))
 		self._update()
 
 	def set_sorting(self, sorting):
@@ -674,9 +947,7 @@ class TagCloudWidget(gtk.TextView):
 		'''Stop the model from listening to the index. Used to unhook
 		the model before reloading the index.
 		'''
-		for id in self._signals:
-			self.index.disconnect(id)
-		self._signals = ()
+		self.disconnect_from(self.index)
 		self._clear()
 		self.index = None
 
@@ -706,7 +977,11 @@ class TagCloudWidget(gtk.TextView):
 	def _update(self, *a):
 		'''Update the cloud to show only tags that share a set of pages
 		with the selected tags.'''
-		selected = [b.indextag for b in self.get_children() if b.get_active()]
+		selected = [b.indextag for b in self.get_children()
+			if b.get_active() and
+			self.index.lookup_tag(b.indextag.name)]
+			# Need the lookup here in case the tag went missing in the
+			# mean time e.g. due to editing of the page
 		self._clear()
 
 		buffer = self.get_buffer()
@@ -756,142 +1031,4 @@ class TagCloudWidget(gtk.TextView):
 gobject.type_register(TagCloudWidget)
 
 
-class TagsPluginWidget(gtk.VPaned):
-	'''Widget combining a tag cloud and a tag based page treeview'''
 
-	def __init__(self, plugin):
-		gtk.VPaned.__init__(self)
-		self.plugin = plugin
-
-		self.plugin.uistate.setdefault('vpane_pos', 150)
-		self.plugin.uistate.setdefault('treeview', 'tagged', set(['tagged', 'tags']))
-		self.plugin.uistate.setdefault('tagcloud_sorting', 'score', set(['alpha', 'score']))
-
-		def add_scrolled(widget):
-			sw = gtk.ScrolledWindow()
-			sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
-			sw.set_shadow_type(gtk.SHADOW_IN)
-			sw.add(widget)
-			self.add(sw)
-			return sw
-
-		self.tagcloud = TagCloudWidget(sorting=self.plugin.uistate['tagcloud_sorting'])
-		add_scrolled(self.tagcloud)
-
-		self.treeview = TagsPageTreeView(self.plugin.ui)
-		add_scrolled(self.treeview)
-
-		self.treeview.connect('populate-popup', self.on_populate_popup)
-		self.tagcloud.connect('selection-changed', self.on_cloud_selection_changed)
-		self.tagcloud.connect('sorting-changed', self.on_cloud_sortin_changed)
-
-		self.plugin.ui.connect('open-page', self.on_open_page)
-		self.plugin.ui.connect('start-index-update', lambda o: self.disconnect_model())
-		self.plugin.ui.connect('end-index-update', lambda o: self.reload_model())
-
-	def finalize_notebook(self, notebook):
-		self.tagcloud.set_sorting(self.plugin.uistate['tagcloud_sorting'])
-
-		self.set_position(self.plugin.uistate['vpane_pos'])
-		def update_uistate(*a):
-			self.plugin.uistate['vpane_pos'] = self.get_position()
-		self.connect('notify::position', update_uistate)
-
-		self.reload_model()
-
-	def on_open_page(self, ui, page, path):
-		self.treeview.select_page(path)
-
-	def toggle_treeview(self):
-		'''Toggle the treeview type in the widget'''
-		if self.plugin.uistate['treeview'] == 'tagged':
-			self.plugin.uistate['treeview'] = 'tags'
-		else:
-			self.plugin.uistate['treeview'] = 'tagged'
-
-		self.reload_model()
-
-	def on_populate_popup(self, treeview, menu):
-		# Add a popup menu item to switch the treeview mode
-		populate_popup_add_separator(menu, prepend=True)
-
-		item = gtk.CheckMenuItem(_('Sort pages by tags')) # T: menu option
-		item.set_active(self.plugin.uistate['treeview'] == 'tags')
-		item.connect_object('toggled', self.__class__.toggle_treeview, self)
-		menu.prepend(item)
-
-		menu.show_all()
-
-	def on_cloud_selection_changed(self, cloud):
-		filter = cloud.get_tag_filter()
-		self.treeview.set_tag_filter(filter)
-
-	def on_cloud_sortin_changed(self, cloud, sorting):
-		self.plugin.uistate['tagcloud_sorting'] = sorting
-
-	def disconnect_model(self):
-		'''Stop the model from listening to the index. Used to
-		unhook the model before reloading the index. Typically
-		should be followed by reload_model().
-		'''
-		self.treeview.disconnect_index()
-		self.tagcloud.disconnect_index()
-
-	def reload_model(self):
-		'''Re-initialize the treeview model. This is called when
-		reloading the index to get rid of out-of-sync model errors
-		without need to close the app first.
-		'''
-		assert self.plugin.uistate['treeview'] in ('tagged', 'tags')
-
-		if self.tagcloud.index is None:
-			self.tagcloud.set_index(self.plugin.ui.notebook.index)
-
-		if self.plugin.uistate['treeview'] == 'tagged':
-			model = TaggedPageTreeStore(self.plugin.ui.notebook.index) # FIXME clean up law of D
-		else: # tags
-			model = TagsPageTreeStore(self.plugin.ui.notebook.index) # FIXME clean up law of D
-
-		self.treeview.set_model(model)
-
-		if self.plugin.ui.page:
-			model.select_page(self.plugin.ui.page)
-
-
-class TagsPlugin(PluginClass):
-
-	plugin_info = {
-		'name': _('Tags'), # T: plugin name
-		'description': _('''\
-This plugin provides a page index filtered by means of selecting tags in a cloud.
-'''), # T: plugin description
-		'author': 'Fabian Moser',
-		'help': 'Plugins:Tags',
-	}
-
-	def __init__(self, ui):
-		PluginClass.__init__(self, ui)
-		self.sidepane_widget = None
-
-	def initialize_ui(self, ui):
-		if self.ui.ui_type == 'gtk':
-			self.connect_embedded_widget()
-
-	def finalize_notebook(self, notebook):
-		if self.ui.ui_type == 'gtk' and self.sidepane_widget:
-			self.sidepane_widget.finalize_notebook(notebook)
-
-	def disconnect(self):
-		self.disconnect_embedded_widget()
-		PluginClass.disconnect(self)
-
-	def connect_embedded_widget(self):
-		if self.sidepane_widget is None:
-			self.sidepane_widget = TagsPluginWidget(self)
-			self.ui.mainwindow.add_tab(_('Tags'), self.sidepane_widget, LEFT_PANE)
-			self.sidepane_widget.show_all()
-
-	def disconnect_embedded_widget(self):
-		if not self.sidepane_widget is None:
-			self.ui.mainwindow.remove(self.sidepane_widget)
-			self.sidepane_widget = None
